#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <sys/file.h>
#include <sys/resource.h>
#include <syslog.h>

#define BASEPORT 33435
#define PORTSPERHOP 3
#define HOPS 30

struct {
   time_t timestamp;
   int addr;
   unsigned short int sport;
   unsigned short int tport;
   unsigned char hop;
} log;

void phelp(char **argv) {
   fprintf(stderr, "%s%s%s%s%s%s%s%s%s%s%u%s%s%u%s%s%u%s",
         "Antiroute 1.1, (c) 1998, 2000, 2001 Neven Lovric <neven@lovric.net>\n",
         "Prevents and logs UDP-based route tracking\n\n",
         "Usage: ", argv[0], " [-o] [-s] [-l logfile] [-d logfile] [-b baseport] [-p portsperhop] [-h hops]\n\n",
         "   -o\tPrint route tracking on stdout\n",
         "   -s\tLog route tracking to the system log\n",
         "   -l\tLog route tracking to <logfile>\n",
         "   -d\tDisplay contents of <logfile>\n",
         "   -b\tBase UDP port used in route tracking [", BASEPORT, "]\n",
         "   -p\tPorts per hop [", PORTSPERHOP, "]\n",
         "   -h\tHops [", HOPS, "]\n\n");
   exit(0);
}

int main (int argc, char **argv) {
   struct sockaddr_in sockaddr;
   int *sock, sockaddrlen, i, displaylogfile = 0, ports, printonstdout = 0, logfilefd = -1, unopened = 0, logtosyslog = 0; // logfilefd - dumb compiler
   char buf[1024], *logfile = NULL, *timestr, *ptr;
   unsigned short int baseport = BASEPORT, portsperhop = PORTSPERHOP, hops = HOPS;
   struct flock lockinfo;
   struct rlimit reslimit;

   sockaddr.sin_family = AF_INET;
   sockaddr.sin_addr.s_addr = 0;
   sockaddrlen = sizeof(struct sockaddr_in);
   
   if (argc < 2)
      phelp(argv);
   for (i = 1; i < argc; i++) {
      if (!strcmp(argv[i], "-o")) {
         printonstdout = 1;
         continue;
      }
      if (!strcmp(argv[i], "-s")) {
         logtosyslog = 1;
         continue;
      }
      if (i == argc - 1)
         phelp(argv);
      if (!strcmp(argv[i], "-l")) {
         logfile = argv[++i];
         continue;
      }
      if (!strcmp(argv[i], "-d")) {
         logfile = argv[++i];
         displaylogfile = 1;
         continue;
      }
      if (!strcmp(argv[i], "-b")) {
         baseport = (unsigned short int)strtoul(argv[++i], NULL, 10);
         continue;
      }
      if (!strcmp(argv[i], "-p")) {
         portsperhop = (unsigned short int)strtoul(argv[++i], NULL, 10);
         continue;
      }
      if (!strcmp(argv[i], "-h")) {
         hops = (unsigned short int)strtoul(argv[++i], NULL, 10);
         continue;
      }
      phelp(argv);
   }
   ports = hops * portsperhop;

   getrlimit(RLIMIT_NOFILE, &reslimit);
   reslimit.rlim_cur =  ports + 4;
   if (setrlimit(RLIMIT_NOFILE, &reslimit) == -1) {
      fprintf(stderr, "Unable to set resource litmit: %s.\n", strerror(errno));
      exit(1);
   }
//   printf("Maximum descriptors: %i\n", getdtablesize());

   if (logfile) {
      logfilefd = open(logfile, displaylogfile ? O_RDONLY : O_CREAT | O_WRONLY, S_IREAD | S_IWRITE);
      if (logfilefd == -1) {
         fprintf(stderr, "Unable to open %s: %s.\n", logfile, strerror(errno));
         exit(1);
      }
      if (!displaylogfile) {
         lockinfo.l_type = F_WRLCK;
         lockinfo.l_whence = SEEK_SET;
         lockinfo.l_start = 0;
         lockinfo.l_len = 0;
         if (fcntl(logfilefd, F_SETLK, &lockinfo) == -1) {
            fprintf(stderr, "Unable to lock %s. Antiroute may already be running.\n", logfile);
            exit(1);
         }
         if (lseek(logfilefd, 0, SEEK_END) == -1) {
            fprintf(stderr, "Unable to reposition file offset on %s: %s.\n", logfile, strerror(errno));
            exit(1);
         }
      }
   }

   if (displaylogfile) {
      while (read(logfilefd, &log, sizeof(log)) == sizeof(log)) {
         timestr = ctime(&log.timestamp);
         if ((ptr = strchr(timestr, '\n')))
            *ptr = 0;
         printf("%s  %u.%u.%u.%u:%u -> port %u (hop %u)\n", timestr, (unsigned char)(log.addr >> 24), (unsigned char)(log.addr >> 16), (unsigned char)(log.addr >> 8), (unsigned char)log.addr, log.sport, log.tport, log.hop);
      }
      exit(0);
   }

   while (!(sock = malloc(ports * sizeof(struct sockaddr_in))))
      usleep(1);
   for (i = 0; i < ports; i++) {
      sock[i] = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
      if (sock[i] == -1) {
         fprintf(stderr, "Socket creation error.\n");
         exit(1);
      }
      sockaddr.sin_port = htons(baseport + i);
      if (bind(sock[i], (struct sockaddr *)&sockaddr, sockaddrlen)) {
         fprintf(stderr, "Unable to opet UDP port %u: %s. Tracking on this port will not be detected.\n", baseport + i, strerror(errno));
         close(sock[i]);
         sock[i] = -1;
         unopened++;
      } else
         fcntl (sock[i], F_SETFL, O_NONBLOCK);
   }
   if (unopened == ports) {
      fprintf(stderr, "Failed to open all selected UDP ports.\n");
      exit(1);
   }

   if (logtosyslog)
      openlog("antiroute", 0, LOG_DAEMON);

   while (1) {
      for (i = 0; i < ports; i++) {
         if (sock[i] == -1)
            continue;
         sockaddr.sin_addr.s_addr = 0;
         log.timestamp = time(NULL);
         if (recvfrom(sock[i], buf, 1024, 0, (struct sockaddr *)&sockaddr, &sockaddrlen) != -1) {
            log.addr = htonl(sockaddr.sin_addr.s_addr);
            log.tport = baseport + i;
            log.hop = i / portsperhop + 1;
            if (printonstdout) {
               timestr = ctime(&log.timestamp);
               if ((ptr = strchr(timestr, '\n')))
                  *ptr = 0;
               printf("%s  %u.%u.%u.%u:%u -> port %u (hop %u)\n", timestr, (unsigned char)(log.addr >> 24), (unsigned char)(log.addr >> 16), (unsigned char)(log.addr >> 8), (unsigned char)log.addr, sockaddr.sin_port, log.tport, log.hop);
            }
            if (logtosyslog)
               syslog(LOG_INFO, "%u.%u.%u.%u:%u -> port %u (hop %u)\n", (unsigned char)(log.addr >> 24), (unsigned char)(log.addr >> 16), (unsigned char)(log.addr >> 8), (unsigned char)log.addr, sockaddr.sin_port, log.tport, log.hop);
            if (logfile) {
               log.sport = sockaddr.sin_port;
               write(logfilefd, &log, sizeof(log));
            }
         }
      }
      usleep(1);
   }
}
